﻿#pragma once
#include "stdlib.h"
#include "stdio.h"
#include "DataTyps.h"
#include "MathExt.h"
#include "QClipboard"
#include "QApplication"

//! official representation of Track in BTP3
/*!
  This class is frequently used by all routing algorithm to create tracks
  from Neighbour array and afterwards modify the Track.\n
  Data is stored in arrays d_h_array and KOOa. Every function modifying the
  Track will realloc these arrays.
*/
class Track
{
public:
    Track(trackpoint* tpl,int tc, KOOlist* kl, int kc);
    //! ordinary way to create a Track: from a list of Neighbour
    Track(Neighbour** narray, int nc);
    //! Track object without height information
    Track(double** KOOarray, int count);
    //! Track object without BTP reference, e.g. when dealing with gpx files
    Track(double *d,double *h,double *lat,double *lon,int count);
    Track(Track* tracktocopy);			//!< deep copy of a Track object
    //! create from BTP track format file
    Track(char* btrfilename, KOO* (idfun)(qint64,void*), void*KOOidSearcher);
    Track();
    ~Track(void);
    void add(trackpoint* tpl,int tc, KOOlist* kl, int kc);
    void add(Track* t2);				//t2 kopieren und anfuegen
    void get_data_ref(trackpoint** tpl,int* tc, KOOlist** kl, int* kc);
    double** get_d_h_array();           //!< returns reference
    double** get_d_h_array_copy();      //!< returns deep copy
    long int get_d_h_length();
    double** get_KOOa();                //!< returns reference
    double** get_KOOa_copy();           //!< returns deep copy
	long int get_KOOa_length();
    float get_height_meters();          //!< calculates total altitude gain
    //! calculates altitude gain on a section of the track
    float get_height_meters(float dmin, float dmax);
    //! return serpentine index
    /*! for each bend calculates bend angle, divides by 180° and squares. Sums
     *  up this value and returns it.
     */
	float get_bend();
	KOO* get_last_KOO();
    KOO* get_first_KOO();
    KOO* get_KOO(int kl_i);
    KOOlist* get_KOOlist();
    trackpoint* get_trackpoint();
    double get_length();                //!< returns Track length [km]
    short notNULL();                    //!< is track empty or not
    void to_gpx(char* filename);        //!< save to gpx file
    void to_binary_KOO(char* filename); //!< save binary KOOa to file (website)
    void to_binaryPlus(char* filename); //!< save binary KOO and Heigh data
    void to_btr(char* filename);        //!< save to BTP track format
    void delete_cross(int nr);          //!< deletes last nr cross of Track
    void delete_to_cross(KOO* k);       //!< deletes cross after k
    void delete_complete();             //!< deletes whole Track content
    //! creates KOO refering distance point d
    short extract_KOO(KOO* k, double d);
    //! creates double for height refering distance point d
	short extract_h(double d, double* h);
    //! searches height of i-th coordinate
    double get_h(int KOO_i);
    //! extract way direction at distance point d
    short extract_azimtut(double d, double* azimut);
    //! searches for closest point on track to point (lat,lon)
    /*! If track is empty return 0, else 1. In d, the the closest point of
     *  track will be stored with its distance on the Track, and NOT the
     *  distance to (lat,lon).
     */
	short next_d(double lat, double lon, double* d);
    //! next_d(), but with left or right orientation
    short next_d_sign(double lat, double lon, double* d, double* dmin0);
    //! stores cross in kdot which is closest to (clickedlat,clickedlon)
	short next_cross(double clickedlat, double clickedlon,
					  KOO** kdot, float maxd);
    //! check if point (lat,lon) is closer to track then limit (degree!)
	short near_check(double lat, double lon, double limit);
    //! increases rect to contain track, if necessary
	void adapt_brect(double* latmin, double* latmax,
					 double* lonmin, double* lonmax);
    //! return bounding rect of Track
	void get_brect(double* latmin, double* latmax,
				   double* lonmin, double* lonmax);
    int owns(KOO* k);                   //! check if k is an element of Track
	void get_anchors(float d, KOO** kfrom, KOO** kto);
	short found_anchors(double clickedlat, double clickedlon,
                        float maxd, int* kl_i_from, int* kl_i_to);
    KOO get_firstKOO();
    KOO get_lastKOO();
    /*Track veraendern*/
	void invert();
	void trim_left(float d);
	void trim_right(float d);
    void trim_left(double d);
    void trim_right(double d);
	void trim_left(KOO* k);
	void trim_right(KOO* k);
    void trim_left(int kl_i);
    void trim_right(int kl_i);
    //! a KOO of the track can be marked or not, used for drag&drop routing
	short toggle_marker(KOO* k);
    void reconstruct_neighbours(Neighbour*** narray, int* nc);


private:
    //! stores dynamically allocated KOO if public a function requires that
    KOO* KOOpool;
    trackpoint* tpl;                    //!< distance height data array
    //! KOO data array
    /*! with marker attribute and cross reference to tpl at every cross*/
    KOOlist* kl;
    int kc;                             //!< array length if kl
    int tc;                             //!< array length if tpl
    double** d_h_array;					//!< distance height output, d:0, h:1
    double** KOOa;						//!< coordinate output, lat:0, lon:1
    void refresh_d_h_array();           //!< refreshs d_h_array
    void refresh_KOOa();                //!< refreshs KOOa
	KOO* add_to_KOOpool(double lat, double lon);
	void free_KOOpool(KOO* k);
    void reset_markers();               //!< set markers to zero
    void trim_left(int kci, int tci);   //!< trim kl to index kci & tpl to tci
    void trim_right(int kci, int tci);  //!< trim kl to index kci & tpl to tci
};

