﻿#ifndef STRONGCALC_H
#define STRONGCALC_H
#include <QThread>
#include "DataTyps.h"
#include "Track.h"
#include "fibunacci.h"
#include "osm.h"

enum RunMode {RMcalc,RMcalcRestricted,RMcalcClimb,RMcalcClimbRestricted,RMload,RMsave};
//! label algorithm to solve multicriteria shortes path problem
/*! This thread is designed to calculated the shortes route at given altitude
 *  difference. According to the original idea to ride the shortest route with
 *  10000 Hm, the solution is call "$STRONG", and everything, which is related
 *  to this calculation will by named alike.
 *  The Label in that case is a STRONG struct, it shows where the track, which
 *  passes the owning cross, comes from, i.g. from which STRONG struct. Every
 *  STRONG struct represents an interval of Hm to make the mathematical problem
 *  solvable at all. Every crossdata bears an array of STRONG structs and
 *  intervals respectivly. The index of such an array, representing a Hm
 *  interval, is called strong index.
 *  In Principle the algorithm works like Dijkstra, with the important
 *  difference, that an ariving track will compare itself with tracks of
 *  similiar Hm having passed the cross before, i.g. tracks which passed the
 *  STRONG struct with the related striong index. In Dijkstra the distance only
 *  counts and tracks in only one STRONG struct are of importance. For further
 *  information refer to multicriteria shortes path problem and label algorithm.
*/
class STRONGcalc: public QThread{
Q_OBJECT
public:
    //! STRONG calculation
    STRONGcalc(OSM* o, KOO* s, KOO* e,char max, char min,
               STRONGsetting* hs,Track*** trck, int* N, bool Hmonly);
    //! save STRONG calculation
    STRONGcalc(OSM* o, int N, char* filename,Track***results);
    //! load STRONG calculation
    STRONGcalc(OSM* o, int*N, char* filename,Track*** results);
    void run();
    void exit();
public slots:
    //! sets STRONGcalc::calcstopped and stops algorithm in next iteration
    void stop_calculations();
signals:
    //! returns succes if a track was extracted from the STRONG network
    void finished(int succesful);
    //! emitted if iteration step has finished
    void refresh_progress(int step);
private:
    int counter;
    //! core function of the algorithm
    /*! takes a STRONG struct from the lowest Hm priority queue. This queue is
     *  a FibunacciHeap. The extracted STRONG is activ and has the shortes
     *  distance. It will be expanded by calc() and compare itself to existing
     *  aspirants in the related STRONG structs of the neighbouring crossdata.
     *  If it has shorter distance or finds a free STRONG struct it will be
     *  kept active and stored in the FibunacciHeap.\n
     *  calc() runs until all FibunacciHeap are empty or calcstopped was set to
     *  one.
     */
    template <class T> void calc(FibunacciHeap<T>**f);
    //!@{
    //! prepares the class for the calc() function
    /*! Calcs maximum strong index, allocates STRONG arrays in all crossdata,
     *  which lie in the specified elliptical area. Allocates the FibunacciHeap
     *  array and fills it with one Element, namely the startpoint.
     */
    void init_Strong();
    void init_StrongPlus();
    void init_StrongClimb();
    //!@}
    //!@{
    //! prepares the class for the calc() function
    /*! Calcs maximum strong index, allocates STRONG arrays in all crossdata,
     *  which lie in the specified elliptical area. Allocates the FibunacciHeap
     *  array and fills it with zero indexed STRONG of all crossdata, which lie
     *  within the data area. This serves as a STRONGcalc with undefined
     *  start point.
     */
    void init_AreaStrong(KOO* k);
    void init_AreaStrongPlus(KOO* k);
    void init_AreaStrongClimb(KOO* k);
    //!@}
    //! allocates STRONG arrays
    /*! allocates STRONG arrays for all crossdata lying within the the
     *  specified data ellipse. This ellipse is defined by start and endpoint
     *  as focal point and the maximum distance from these points, stored in
     *  radius variable.
     */
    void setSTRONG(KOO* k);
    //! reconstructs track and checks if Neighbour n was already used
    /*! If the Neighbour was found, zero is returned, else one*/
    short traceback(STRONG* s, Neighbour* n);
    //!@{
    //! calculates related Hm intervall from Hm, ig.g strong index
    int index_STRONG(STRONGfib* t,Neighbour* n);
    int index_STRONG(STRONGPlusfib* t,Neighbour* n);
    int index_STRONG(STRONGClimbfib* t,Neighbour* n);
    //!@}
    //!@{
    //! stores succesfully expanded STRONG in the FibunacciHeap
    void store_STRONG(STRONGfib* t, Neighbour* n,int si, int i);
    void store_STRONG(STRONGPlusfib* t, Neighbour* n,int si, int i);
    void store_STRONG(STRONGClimbfib* t, Neighbour* n,int si, int i);
    //!@}
    //! reconstructs strong track with strong index i
    /*! This is tracing back from end point or if no end point was stated then
     *  the point which was reached with least distance.
     */
    void STRONG_to_trck(int i);
    //! reconstructs link between start STRONG and end STRONG
    /*! generates a list of used Neighbour which can be uses to create a Track
     *  object
     */
    void extract_neighbourlist(STRONG* s, STRONG* e,
                               Neighbour ***nl,long int* nc);
    //! checks if STRONG can expand concerning restrictions
    int extract_STRONGPlus_restriciton(Neighbour* n);
    //! checks if STRONG can expand concerning restrictions
    int extract_STRONGClimb_restriciton(Neighbour* n);
    //! executes the loading of a STRONG network from file filename
    void load_STRONG();
    //! executes the save of a STRONG network to file filename
    void save_STRONG();
    //! low level load routine of load_STRONG
    void load_crossdata(FILE* f);
    //! low level save routine of save_STRONG
    void save_crossdata(KOO *k, FILE* f);
    RunMode runmode;                    //!< type of STRONGcalc to be executed
    OSM* o;                             //!< dataset to operate on
    char filename[999];                 //!< for save and load
    KOO *s;                             //!< start point
    KOO *e;                             //!< end point
    char max, min;                      //!< restrictions on Neighbour Way type
    STRONGsetting* hs;                  //!< major settings of the calculation
    Track*** trck;                      //!< intermediate results of iterations
    /*Variablen des STRONG-Algorithmus*/
    int N, *Nout, calcstopped;
    //! data ellipse radius calculated from STRONGsetting and start/ end point
    float radius;
    short foundSTRONG;                  //!< if 1 return succes when finished
    Neighbour *dummy;                   //!< bumbper to mark root if STRONG net
    //!@{
    //! priority queue of the calc() routine
    FibunacciHeap<STRONGfib>** f;
    FibunacciHeap<STRONGPlusfib>** fplus;
    FibunacciHeap<STRONGClimbfib>** fclimb;
    //!@}
    double* mind;                       //!< minimum distance for strong index
    //! end point array of shortest track at specific strong index
    STRONG** mindSTRONG;
    //!@{
    //! data of the currently traded STRONG struct
    float hm, hmup, hmdown, Psum, dP;
    //!@}
};

#endif // STRONGCALC_H
